﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;

namespace K4W_v2_Depth
{
    class DistanceEstimation
    {
        //*****************************************************
        //【基底遷移アルゴリズム】『放物面推定』
        //   O(xj,yj)=xj*(axj+b)+yj*(cyj+d)+e
        //*****************************************************
        public struct Parabora
        {
            public int Xs;  //奇数である事
            public int Ys;  //奇数である事
            public int Xs2; // Xspan/2
            public int Ys2; // Yspan/2
            public int n;   //サンプル数
            public ushort[] S; //【サンプル】S(Xj,Yj)
            public int[] X; // Xj
            public int[] Y; // Yj
            public double[] K; // Ka, Kb, Kc, Kd, Ke
            public double[] U; // a, b, c, d, e (未知数)
        }

        public Parabora P;

        //*****************************************************
        //【基底遷移アルゴリズム】『放物面推定』page初期化時
        //*****************************************************
        public void InitializeBaseTransition(ushort[] Buf, int Xsize, Rectangle R)
        {
            // XsizeはデータバッファBufの水平方向画素数
            // 領域Rの横幅と高さは必ず奇数でなければならない。
            double x2, y2; int j, XX, YY, Xj, Yj; ushort d;
            double[] sum = new double[5]; //【係数分母】
            P.Xs = R.Width;               //【領域横幅】
            P.Ys = R.Height;              //【領域高さ】
            P.Xs2 = (int)(P.Xs / 2);      // -Xs/2 ～ +Xs/2
            P.Ys2 = (int)(P.Ys / 2);      // -Ys/2 ～ +Ys/2
            P.n = R.Width * R.Height;     //【サンプル数n】
            P.S = new ushort[P.n];        //【サンプル配列S(xj,yj)確保】
            P.X = new int[P.n];           //【サンプル配列S(xj,yj)確保】
            P.Y = new int[P.n];           //【サンプル配列S(xj,yj)確保】
            P.K = new double[5];          // Ka, Kb, Kc, Kd, Ke
            P.U = new double[5];          // a, b, c, d, e
            // サンプル値取得 S(xj,Yj)
            j = 0;
            Yj = -P.Ys2;
            for (YY = R.Y; YY < R.Y + R.Height; YY++)
            {
                Xj = -P.Xs2;
                for (XX = R.X; XX < R.X + R.Width; XX++)
                {
                    d = Buf[XX + Xsize * YY];
                    if((d>500) && (d<8000)) { P.S[j] = d; P.X[j] = Xj; P.Y[j] = Yj; j++; }
                    Xj++;
                }
                Yj++;
            }
            P.n = j; // 実際に有効であったサンプルの数
            // 乗算係数 Ka, Kb, Kc, Kd, Keの計算
            sum[4] = P.n;
            for (j = 0; j < 4; j++) { sum[j] = 0.0; } //【係数分母の初期化】
            for(j = 0; j < P.n; j++)
            {
                x2 = (double)P.X[j]; x2 = x2 * x2;
                y2 = (double)P.Y[j]; y2 = y2 * y2;
                sum[0] += x2 * x2;
                sum[1] += x2;
                sum[2] += y2 * y2;
                sum[3] += y2;
            }
            for (j = 0; j < sum.Length; j++) { P.K[j] = 1.0 / sum[j]; } //【係数演算】
            // 未知数初期化
            for (j = 0; j < P.U.Length; j++) { P.U[j] = 0.0; }
        }

        //*****************************************************
        //【単純平均アルゴリズム】
        //*****************************************************
        public double MeasureRangeAverageDistance()
        {
            double sum = 0.0;
            for (int j = 0; j < P.n; j++) { sum += (double)P.S[j]; }
            return sum / (double)(P.n);
        }

        //*****************************************************
        //【基底遷移アルゴリズム】『放物面推定』
        //*****************************************************
        public double BaseTransition()
        {
            //【未知数推定】漸化式による推定を10回繰り返す
            int i, j;
            double x, y, sum;
            for (i = 0; i < 10; i++) // iteration
            {
                //【未知数①】a: P.U[0]
                sum = 0.0;
                for ( j = 0; j < P.n; j++)
                {
                    x = (double)P.X[j]; y = (double)P.Y[j];
                    sum += x * x * (P.S[j] - ((P.U[0] * x + P.U[1]) * x + (P.U[2] * y + P.U[3]) * y + P.U[4]));
                }
                P.U[0] += P.K[0] * sum;

                //【未知数②】b: P.U[1]
                sum = 0.0;
                for (j = 0; j < P.n; j++)
                {
                    x = (double)P.X[j]; y = (double)P.Y[j];
                    sum += x * (P.S[j] - ((P.U[0] * x + P.U[1]) * x + (P.U[2] * y + P.U[3]) * y + P.U[4]));
                }
                P.U[1] += P.K[1] * sum;

                //【未知数③】c: P.U[2]
                sum = 0.0;
                for (j = 0; j < P.n; j++)
                {
                    x = (double)P.X[j]; y = (double)P.Y[j];
                    sum += y * y * (P.S[j] - ((P.U[0] * x + P.U[1]) * x + (P.U[2] * y + P.U[3]) * y + P.U[4]));
                }
                P.U[2] += P.K[2] * sum;

                //【未知数④】d: P.U[3]
                sum = 0.0;
                for (j = 0; j < P.n; j++)
                {
                    x = (double)P.X[j]; y = (double)P.Y[j];
                    sum += y * (P.S[j] - ((P.U[0] * x + P.U[1]) * x + (P.U[2] * y + P.U[3]) * y + P.U[4]));
                }
                P.U[3] += P.K[3] * sum;

                //【未知数⑤】e: P.U[4]
                sum = 0.0;
                for (x = y = j = 0; j < P.n; j++)
                {
                    x = (double)P.X[j]; y = (double)P.Y[j];
                    sum += (P.S[j] - ((P.U[0] * x + P.U[1]) * x + (P.U[2] * y + P.U[3]) * y + P.U[4]));
                }
                P.U[4] += P.K[4] * sum;
            }

            //【放物面未知数】
            //P.Xc = -P.U[1] / (2.0 * P.U[0]);
            //P.Yc = -P.U[3] / (2.0 * P.U[2]);
            //P.Zc = ((P.U[0] * P.X + P.U[1]) * P.X + (P.U[2] * P.Y + P.U[3]) * P.Y + P.U[4]) * 1000.0;
            //P.Xc = P.Xc * 1000.0;
            //P.Yc = P.Yc * 1000.0;

            return P.U[4]; // 領域中央の画素の距離
        }
    }
}
